//
//  DevicesController.m
//  HIDController
//
//  Created by Joel Jordan on 11/30/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/IOMessage.h>
#include <IOKit/hid/IOHIDLib.h>
#include <IOKit/hid/IOHIDKeys.h>
#include <IOKit/hid/IOHIDUsageTables.h>
#include <IOKit/hidsystem/IOHIDLib.h>
#include <IOKit/hidsystem/IOHIDShared.h>
#include <IOKit/hidsystem/IOHIDParameter.h>

#import "DevicesController.h"
#import "HIDDevice.h"

void deviceAddedCallback(void *refcon, io_iterator_t devIterator) {
	DevicesController *this;
	io_object_t device;
	
	this = (DevicesController *) refcon;

	while (device = IOIteratorNext(devIterator)) {
		HIDDevice *hidDev;
		
		hidDev = [[HIDDevice alloc] initWithDeviceObject:device];
		[[this deviceController] addObject:hidDev];
		[hidDev release];
	}
}

void deviceRemovedCallback(void *refcon, io_iterator_t devIterator) {
	DevicesController *this;
	io_object_t device;
	
	this = (DevicesController *) refcon;
	
	while (device = IOIteratorNext(devIterator)) {
		// find the HID Device this corresponds to
		HIDDevice *hidDev;
		NSEnumerator *devEnum;
		
		devEnum = [[this devices] objectEnumerator];
		while (hidDev = [devEnum nextObject]) {
			if (IOObjectIsEqualTo(device, [hidDev devObj])) {
				[[this deviceController] removeObject:hidDev];
				break;
			}
		}
	}
}


@implementation DevicesController
- (id)init {
	
	self = [super init];
	if (!self) {
		return self;
	}
	
	devices = [[NSMutableArray alloc] init];
	return self;
}

- (void)awakeFromNib
{
	kern_return_t kr;
	io_master_t masterPort;
	NSMutableDictionary *matchingDict;
	
	// create master port
	kr = IOMasterPort(kIOMasterPortDefault,&masterPort);
	if (KERN_SUCCESS != kr) {
		NSException* myException = [NSException
        exceptionWithName:@"IOException"
				   reason:@"Could not create master port"
				 userInfo:nil];
		@throw myException;
	}
	
	// get the matching dictionary for HID devices
	
	matchingDict = (NSMutableDictionary *) IOServiceMatching(kIOHIDDeviceKey);
	[matchingDict retain]; //we need two copies of this
	
	// set up device insertion notifications
	notificationObject = IONotificationPortCreate(masterPort);
	
	notificationRunLoopSource = IONotificationPortGetRunLoopSource(notificationObject);
	
	CFRunLoopAddSource(CFRunLoopGetCurrent(), notificationRunLoopSource, kCFRunLoopDefaultMode);
	
	kr = IOServiceAddMatchingNotification(notificationObject,kIOMatchedNotification, (CFDictionaryRef) matchingDict, deviceAddedCallback, (void *)self, &addedIterator);

	if (KERN_SUCCESS != kr) {
		NSException* myException = [NSException
        exceptionWithName:@"IOException"
				   reason:@"Could not create device addition notifications"
				 userInfo:nil];
		@throw myException;
	}

	deviceAddedCallback(self, addedIterator);

	kr = IOServiceAddMatchingNotification(notificationObject,kIOTerminatedNotification, (CFDictionaryRef) matchingDict, deviceRemovedCallback, (void *)self, &removedIterator);
	
	if (KERN_SUCCESS != kr) {
		NSException* myException = [NSException
        exceptionWithName:@"IOException"
				   reason:@"Could not create device removal notifications"
				 userInfo:nil];
		@throw myException;
	}
	
	deviceRemovedCallback(self, removedIterator);

}

					   
- (NSMutableArray *)devices {
	return devices;
}

- (void)addDevice:(HIDDevice *)newDevice {
	[devices addObject:newDevice];
}

- (NSArrayController *)deviceController {
	return deviceController;
}

- (void)dealloc {
	[devices release];
	IONotificationPortDestroy(notificationObject);
	[super dealloc];
}

@end
