//
//  HIDDevice.m
//  HIDController
//
//  Created by Joel Jordan on 11/28/07.
//  Copyright 2007 L510 Labs. All rights reserved.
//

#import "HIDDevice.h"
#import "HIDElement.h"

static void elementChangedCallback(void *target, IOReturn result, void *refcon, void *sender) {
	// remove the item from the queue
	AbsoluteTime zeroTime = {0,0};
	HIDDevice *hidDev;
	IOHIDQueueInterface **queue;
	IOHIDEventStruct event;
	
	hidDev = (HIDDevice *)target;
	queue = (IOHIDQueueInterface **)sender;
	
//	NSLog(@"elementChangedCallback entered for device %@", [hidDev productName]);
	
	if (!hidDev || (queue != [hidDev queue])) {
		return;
	}
	
	while (result == kIOReturnSuccess) {
		result = (*queue)->getNextEvent(queue, &event, zeroTime, 0);
		if (result != kIOReturnSuccess) {
			continue;
		}
		if ((event.longValueSize != 0) && (event.longValue != NULL)) {
			free(event.longValue);
			continue;
		}

		// look up the element and make it active
		
		HIDElement *element;
		element = [[hidDev cookieDict] objectForKey:[NSNumber numberWithInt:(int)event.elementCookie]];
		
		[hidDev addActiveElement:element];

		// FIXME: set up a single timer for the application that's created/invalidated as needed
		
		NSTimer *timer;
		timer = [NSTimer scheduledTimerWithTimeInterval:0.2 target:hidDev selector:@selector(elementNoLongerActive:)  userInfo:element repeats:NO];
		
	}	
		
}

static BOOL addElementsToQueue(IOHIDQueueInterface **queue, NSArray *elements) {
	NSEnumerator *elementEnum;
	IOReturn	 result;
	
	elementEnum = [elements objectEnumerator];
	HIDElement *hidElement;

	while (hidElement = [elementEnum nextObject]) {
		// result will only be success on leaf elements
		// so ignore the return code now and always return TRUE
		result = (*queue)->addElement(queue, [hidElement cookie], 0);
		//NSLog(@"adding element to the queue: %@, result = %ul", [hidElement name], result);
		
		if ([[hidElement elements] count] > 0) {
			addElementsToQueue(queue, [hidElement elements]);
		}
		
		/*		if (result != kIOReturnSuccess) {
			(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
				exceptionWithName:@"IOException"
						   reason:@"Error adding element to queue"
						 userInfo:nil];
		@throw myException;
		}
		*/
	}
	return YES;
}	

static void makeCookieDictionary(NSArray *elements, NSMutableDictionary *cookieDict) {
	NSEnumerator *elementEnum;
	HIDElement *hidElement;
	if ([elements count]) {
		elementEnum = [elements objectEnumerator];
		while (hidElement = [elementEnum nextObject]) {
			[cookieDict setObject:hidElement forKey:[NSNumber numberWithInt:((int)[hidElement cookie])]];
			makeCookieDictionary([hidElement elements], cookieDict);
		}
	}
}

@implementation HIDDevice

- (id)initWithDeviceObject:(io_service_t)deviceObject
{
	IOReturn			result;
	NSMutableDictionary *properties;
	SInt32				score = 0;
	IOCFPlugInInterface **plugInInterface = NULL;
	[super init];
	
	hidDev = NULL;	
	queue = NULL;
	elementActive = NO;
	devObj = deviceObject;
	cookieDict = [[NSMutableDictionary alloc] init];
	activeElements = [[NSMutableArray alloc] init];
	
	// first get some information about the object from its properties dictionary
	
	result = IORegistryEntryCreateCFProperties(deviceObject, (CFMutableDictionaryRef *) &properties,
											   kCFAllocatorDefault, kNilOptions);
	
	vendorID = (NSNumber *) [properties objectForKey:@kIOHIDVendorIDKey];
	[vendorID retain];
	productID = (NSNumber *) [properties objectForKey:@kIOHIDProductIDKey];
	[productID retain];
	
	transport = (NSString *) [properties objectForKey:@kIOHIDTransportKey];
	[transport retain];
	versionNumber = (NSNumber *) [properties objectForKey:@kIOHIDVersionNumberKey];
	[versionNumber retain];
	manufacturer = (NSString *) [properties objectForKey:@kIOHIDManufacturerKey];
	[manufacturer retain];
	productName = (NSString *) [properties objectForKey:@kIOHIDProductKey];
	[productName retain];
	serialNumber = (NSString *) [properties objectForKey:@kIOHIDSerialNumberKey];
	[serialNumber retain];
	
	// find the device elements
	// usually, there's an array off the device root
	NSMutableArray *newElements = [[NSMutableArray alloc] init];
	NSEnumerator *elementEnum = [[properties objectForKey:@kIOHIDElementKey] objectEnumerator];
	NSDictionary *elementDict;
	while (elementDict = [elementEnum nextObject]) {
		HIDElement *newElement = [[HIDElement alloc] initFromDictionary:elementDict];
		[newElements addObject:newElement];
	}
	elements = [newElements copy];
	[newElements release];
	[properties release];
	
	makeCookieDictionary(elements, cookieDict);

	// set up event queues for the elements
	
	// FIXME: find a graceful way to handle all these ridiculous exceptions 
	result = IOCreatePlugInInterfaceForService(devObj, kIOHIDDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
	
	if (result != kIOReturnSuccess) {
		NSException* myException = [NSException
        exceptionWithName:@"IOException"
				   reason:@"Could not create device interface"
				 userInfo:nil];
		@throw myException;
	}
	
	result = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID), (LPVOID) &hidDev);
	
	if (result != S_OK) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error getting device interface"
					 userInfo:nil];
		@throw myException;
	}
	
	// open device
	result = (*hidDev)->open(hidDev, 0);
	if (result != kIOReturnSuccess) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error opening device"
					 userInfo:nil];
		@throw myException;
	}

	// allocate a queue for events on this device
	queue = (*hidDev)->allocQueue(hidDev);
	
	if (!queue) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
		exceptionWithName:@"IOException"
				   reason:@"Error allocating queue"
				 userInfo:nil];
		@throw myException;
	}
	result = (*queue)->create(queue, 0, 8);
	
	if (result != kIOReturnSuccess) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error creating queue"
					 userInfo:nil];
		@throw myException;
	}		
	
	// add all the elements to the queue
	if (!addElementsToQueue(queue, elements)) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
		exceptionWithName:@"IOException"
				   reason:@"Error creating queue"
				 userInfo:nil];
		@throw myException;
	}		
		
	// now enable the callback function
	CFRunLoopSourceRef eventSource;
	
	result = (*queue)->createAsyncEventSource(queue, &eventSource);
	if (kIOReturnSuccess != result) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error creating asynchronous event source"
					 userInfo:nil];
		@throw myException;
	}		

	result = (*queue)->setEventCallout(queue, elementChangedCallback, self, NULL);
	if (kIOReturnSuccess != result) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error setting event callback"
					 userInfo:nil];
		@throw myException;
	}		
	
	CFRunLoopAddSource(CFRunLoopGetCurrent(), eventSource, kCFRunLoopDefaultMode);
					
	result = (*queue)->start(queue);
	
	if (kIOReturnSuccess != result) {
		(*hidDev)->Release(hidDev);
		hidDev = NULL;
		NSException* myException = [NSException
			exceptionWithName:@"IOException"
					   reason:@"Error starting the queue"
					 userInfo:nil];
		@throw myException;
	}		
	
	
	return self;
}	

- (NSString *)transport {
	return transport;
}

- (NSNumber *)vendorID {
	return vendorID;
}

- (NSNumber *)productID {
	return productID;
}

- (NSNumber *)versionNumber {
	return versionNumber;
}

- (NSString *)manufacturer {
	return manufacturer;
}

- (NSString *)productName {
	return productName;
}

- (NSString *)serialNumber {
	return serialNumber;
}

- (NSArray *)elements {
	return elements;
}

- (io_service_t) devObj {
	return devObj;
}

- (BOOL)elementActive {
	return elementActive;
}

- (void)setElementActive:(BOOL)newElementActive {
	elementActive = newElementActive;
}

- (IOHIDQueueInterface **)queue {
	return queue;
}

- (void)addActiveElement:(HIDElement *)element {
	// there's a race condition here, probably not an issue
	if (0 == [activeElements count]) {
		[self setElementActive:YES];
	}
	[activeElements addObject:element];
	[element setActive:YES];
//	NSLog(@"addActiveElement count: %d", [activeElements count]);
}

- (void)removeActiveElement:(HIDElement *)element {
	// there's a race condition here, probably not an issue
	[element setActive:NO];
	[activeElements removeObjectAtIndex:[activeElements indexOfObject:element]];
	if (0 == [activeElements count]) {
		[self setElementActive:NO];
	}
//	NSLog(@"removeActiveElement count: %d", [activeElements count]);
}

- (void)elementNoLongerActive:(NSTimer *)timer {
	HIDElement *element = (HIDElement *)[timer userInfo];
	// decrease active element count
	[self removeActiveElement:element];
}

- (NSDictionary *)cookieDict {
	return cookieDict;
}

- (void)dealloc {
	if (queue) {
		(*queue)->stop(queue);
		(*queue)->Release(queue);
	}
	if (hidDev) {
		(*hidDev)->close(hidDev);
		(*hidDev)->Release(hidDev);
	}
	[activeElements release];
	[cookieDict release];
	[vendorID release];
	[productID release];
	[transport release];
	[versionNumber release];
	[manufacturer release];
	[productName release];
	[serialNumber release];
	[elements release];
	[super dealloc];
}

@end
